import re
from enum import Enum
from typing import Tuple, List, TextIO, Optional, Dict, Set

import click
import functools
import validators
from pprint import pprint
from kgx.config import get_jsonld_context, get_logger
from kgx.graph.base_graph import BaseGraph
from kgx.utils.kgx_utils import (
    get_toolkit,
    snakecase_to_sentencecase,
    sentencecase_to_snakecase,
    camelcase_to_sentencecase,
)
from kgx.prefix_manager import PrefixManager

log = get_logger()


class ErrorType(Enum):
    """
    Validation error types
    """

    MISSING_NODE_PROPERTY = 1
    MISSING_EDGE_PROPERTY = 2
    INVALID_NODE_PROPERTY_VALUE_TYPE = 3
    INVALID_NODE_PROPERTY_VALUE = 4
    INVALID_EDGE_PROPERTY_VALUE_TYPE = 5
    INVALID_EDGE_PROPERTY_VALUE = 6
    NO_CATEGORY = 7
    INVALID_CATEGORY = 8
    NO_EDGE_PREDICATE = 9
    INVALID_EDGE_PREDICATE = 10


class MessageLevel(Enum):
    """
    Message level for validation reports
    """

    # Recommendations
    INFO = 1
    # Message to convey 'should'
    WARNING = 2
    # Message to convey 'must'
    ERROR = 3


class ValidationError(object):
    """
    ValidationError class that represents an error.

    Parameters
    ----------
    entity: str
        The node or edge entity that is failing validation
    error_type: kgx.validator.ErrorType
        The nature of the error
    message: str
        The error message
    message_level: kgx.validator.MessageLevel
        The message level

    """

    def __init__(
        self, entity: str, error_type: ErrorType, message: str, message_level: MessageLevel
    ):
        self.entity = entity
        self.error_type = error_type
        self.message = message
        self.message_level = message_level

    def __str__(self):
        return f"[{self.message_level.name}][{self.error_type.name}] {self.entity} - {self.message}"

    def as_dict(self):
        return {
            'entity': self.entity,
            'error_type': self.error_type.name,
            'message': self.message,
            'message_level': self.message_level.name,
        }


class Validator(object):
    """
    Class for validating a property graph.

    Parameters
    ----------
    verbose: bool
        Whether the generated report should be verbose or not (default: ``False``)

    """

    def __init__(self, verbose: bool = False):
        self.toolkit = get_toolkit()
        self.prefix_manager = PrefixManager()
        self.jsonld = get_jsonld_context()
        self.prefixes = Validator.get_all_prefixes(self.jsonld)
        self.required_node_properties = Validator.get_required_node_properties()
        self.required_edge_properties = Validator.get_required_edge_properties()
        self.verbose = verbose

    @staticmethod
    def get_all_prefixes(jsonld: Optional[Dict] = None) -> set:
        """
        Get all prefixes from Biolink Model JSON-LD context.

        It also sets ``self.prefixes`` for subsequent access.

        Parameters
        ---------
        jsonld: Optional[Dict]
            The JSON-LD context

        Returns
        -------
        Optional[Dict]
            A set of prefixes

        """
        if not jsonld:
            jsonld = get_jsonld_context()
        prefixes = set(jsonld.keys())
        if 'biolink' not in prefixes:
            prefixes.add('biolink')
        return prefixes

    @staticmethod
    def get_required_node_properties() -> list:
        """
        Get all properties for a node that are required, as defined by Biolink Model.

        Returns
        -------
        list
            A list of required node properties

        """
        toolkit = get_toolkit()
        node_properties = toolkit.get_all_node_properties()
        required_properties = []
        for p in node_properties:
            element = toolkit.get_element(p)
            if element and element.deprecated is None:
                if hasattr(element, 'required') and element.required:
                    formatted_name = sentencecase_to_snakecase(element.name)
                    required_properties.append(formatted_name)
                elif element.name == 'category':
                    formatted_name = sentencecase_to_snakecase(element.name)
                    required_properties.append(formatted_name)
        return required_properties

    @staticmethod
    def get_required_edge_properties() -> list:
        """
        Get all properties for an edge that are required, as defined by Biolink Model.

        Returns
        -------
        list
            A list of required edge properties

        """
        toolkit = get_toolkit()
        edge_properties = toolkit.get_all_edge_properties()
        required_properties = []
        for p in edge_properties:
            element = toolkit.get_element(p)
            if element and element.deprecated is None:
                if hasattr(element, 'required') and element.required:
                    formatted_name = sentencecase_to_snakecase(element.name)
                    required_properties.append(formatted_name)
        return required_properties

    def validate(self, graph: BaseGraph) -> list:
        """
        Validate nodes and edges in a graph.
        TODO: Support strict mode

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to validate

        Returns
        -------
        list
            A list of errors for a given graph

        """
        node_errors = self.validate_nodes(graph)
        edge_errors = self.validate_edges(graph)
        return node_errors + edge_errors

    def validate_nodes(self, graph: BaseGraph) -> list:
        """
        Validate all the nodes in a graph.

        This method validates for the following,
        - Node properties
        - Node property type
        - Node property value type
        - Node categories

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to validate

        Returns
        -------
        list
            A list of errors for a given graph

        """
        errors = []
        with click.progressbar(graph.nodes(data=True), label='Validating nodes in graph') as bar:
            for n, data in bar:
                e1 = Validator.validate_node_properties(n, data, self.required_node_properties)
                e2 = Validator.validate_node_property_types(n, data)
                e3 = Validator.validate_node_property_values(n, data)
                e4 = Validator.validate_categories(n, data)
                errors += e1 + e2 + e3 + e4
        return errors

    def validate_edges(self, graph: BaseGraph) -> list:
        """
        Validate all the edges in a graph.

        This method validates for the following,
        - Edge properties
        - Edge property type
        - Edge property value type
        - Edge label

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to validate

        Returns
        -------
        list
            A list of errors for a given graph

        """
        errors = []
        with click.progressbar(graph.edges(data=True), label='Validate edges in graph') as bar:
            for u, v, data in bar:
                e1 = Validator.validate_edge_properties(u, v, data, self.required_edge_properties)
                e2 = Validator.validate_edge_property_types(u, v, data)
                e3 = Validator.validate_edge_property_values(u, v, data)
                e4 = Validator.validate_edge_predicate(u, v, data)
                errors += e1 + e2 + e3 + e4
        return errors

    @staticmethod
    def validate_node_properties(node: str, data: dict, required_properties: list) -> list:
        """
        Checks if all the required node properties exist for a given node.

        Parameters
        ----------
        node: str
            Node identifier
        data: dict
            Node properties
        required_properties: list
            Required node properties

        Returns
        -------
        list
            A list of errors for a given node

        """
        errors = []
        for p in required_properties:
            if p not in data:
                error_type = ErrorType.MISSING_NODE_PROPERTY
                message = f"Required node property '{p}' missing"
                errors.append(ValidationError(node, error_type, message, MessageLevel.ERROR))
        return errors

    @staticmethod
    def validate_edge_properties(
        subject: str, object: str, data: dict, required_properties: list
    ) -> list:
        """
        Checks if all the required edge properties exist for a given edge.

        Parameters
        ----------
        subject: str
            Subject identifier
        object: str
            Object identifier
        data: dict
            Edge properties
        required_properties: list
            Required edge properties

        Returns
        -------
        list
            A list of errors for a given edge

        """
        errors = []
        for p in required_properties:
            if p not in data:
                if p == 'association_id':
                    # check for 'id' property instead
                    if 'id' not in data:
                        error_type = ErrorType.MISSING_EDGE_PROPERTY
                        message = f"Required edge property '{p}' missing"
                        errors.append(
                            ValidationError(
                                f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                            )
                        )
                else:
                    error_type = ErrorType.MISSING_EDGE_PROPERTY
                    message = f"Required edge property '{p}' missing"
                    errors.append(
                        ValidationError(
                            f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                        )
                    )
        return errors

    @staticmethod
    def validate_node_property_types(node: str, data: dict) -> list:
        """
        Checks if node properties have the expected value type.

        Parameters
        ----------
        node: str
            Node identifier
        data: dict
            Node properties

        Returns
        -------
        list
            A list of errors for a given node

        """
        toolkit = get_toolkit()
        errors = []
        error_type = ErrorType.INVALID_NODE_PROPERTY_VALUE_TYPE
        if not isinstance(node, str):
            message = "Node property 'id' expected to be of type 'string'"
            errors.append(ValidationError(node, error_type, message, MessageLevel.ERROR))

        for key, value in data.items():
            element = toolkit.get_element(key)
            if element:
                if hasattr(element, 'typeof'):
                    if element.typeof == 'string' and not isinstance(value, str):
                        message = f"Node property '{key}' expected to be of type '{element.typeof}'"
                        errors.append(
                            ValidationError(node, error_type, message, MessageLevel.ERROR)
                        )
                    elif (
                        element.typeof == 'uriorcurie'
                        and not isinstance(value, str)
                        and not validators.url(value)
                    ):
                        message = f"Node property '{key}' expected to be of type 'uri' or 'CURIE'"
                        errors.append(
                            ValidationError(node, error_type, message, MessageLevel.ERROR)
                        )
                    elif element.typeof == 'double' and not isinstance(value, (int, float)):
                        message = f"Node property '{key}' expected to be of type '{element.typeof}'"
                        errors.append(
                            ValidationError(node, error_type, message, MessageLevel.ERROR)
                        )
                    else:
                        log.warning(
                            "Skipping validation for Node property '{}'. Expected type '{}' vs Actual type '{}'".format(
                                key, element.typeof, type(value)
                            )
                        )
                if hasattr(element, 'multivalued'):
                    if element.multivalued:
                        if not isinstance(value, list):
                            message = f"Multi-valued node property '{key}' expected to be of type '{list}'"
                            errors.append(
                                ValidationError(node, error_type, message, MessageLevel.ERROR)
                            )
                    else:
                        if isinstance(value, (list, set, tuple)):
                            message = f"Single-valued node property '{key}' expected to be of type '{str}'"
                            errors.append(
                                ValidationError(node, error_type, message, MessageLevel.ERROR)
                            )
        return errors

    @staticmethod
    def validate_edge_property_types(subject: str, object: str, data: dict) -> list:
        """
        Checks if edge properties have the expected value type.

        Parameters
        ----------
        subject: str
            Subject identifier
        object: str
            Object identifier
        data: dict
            Edge properties

        Returns
        -------
        list
            A list of errors for a given edge

        """
        toolkit = get_toolkit()
        errors = []
        error_type = ErrorType.INVALID_EDGE_PROPERTY_VALUE_TYPE
        if not isinstance(subject, str):
            message = "'subject' of an edge expected to be of type 'string'"
            errors.append(
                ValidationError(f"{subject}-{object}", error_type, message, MessageLevel.ERROR)
            )
        if not isinstance(object, str):
            message = "'object' of an edge expected to be of type 'string'"
            errors.append(
                ValidationError(f"{subject}-{object}", error_type, message, MessageLevel.ERROR)
            )

        for key, value in data.items():
            element = toolkit.get_element(key)
            if element:
                if hasattr(element, 'typeof'):
                    if element.typeof == 'string' and not isinstance(value, str):
                        message = f"Edge property '{key}' expected to be of type 'string'"
                        errors.append(
                            ValidationError(
                                f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                            )
                        )
                    elif (
                        element.typeof == 'uriorcurie'
                        and not isinstance(value, str)
                        and not validators.url(value)
                    ):
                        message = f"Edge property '{key}' expected to be of type 'uri' or 'CURIE'"
                        errors.append(
                            ValidationError(
                                f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                            )
                        )
                    elif element.typeof == 'double' and not isinstance(value, (int, float)):
                        message = f"Edge property '{key}' expected to be of type 'double'"
                        errors.append(
                            ValidationError(
                                f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                            )
                        )
                    else:
                        log.warning(
                            "Skipping validation for Edge property '{}'. Expected type '{}' vs Actual type '{}'".format(
                                key, element.typeof, type(value)
                            )
                        )
                if hasattr(element, 'multivalued'):
                    if element.multivalued:
                        if not isinstance(value, list):
                            message = (
                                f"Multi-valued edge property '{key}' expected to be of type 'list'"
                            )
                            errors.append(
                                ValidationError(
                                    f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                                )
                            )
                    else:
                        if isinstance(value, (list, set, tuple)):
                            message = (
                                f"Single-valued edge property '{key}' expected to be of type 'str'"
                            )
                            errors.append(
                                ValidationError(
                                    f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                                )
                            )
        return errors

    @staticmethod
    def validate_node_property_values(node: str, data: dict) -> list:
        """
        Validate a node property's value.

        Parameters
        ----------
        node: str
            Node identifier
        data: dict
            Node properties

        Returns
        -------
        list
            A list of errors for a given node

        """
        errors = []
        error_type = ErrorType.INVALID_NODE_PROPERTY_VALUE
        if not PrefixManager.is_curie(node):
            message = f"Node property 'id' expected to be of type 'CURIE'"
            errors.append(ValidationError(node, error_type, message, MessageLevel.ERROR))
        else:
            prefix = PrefixManager.get_prefix(node)
            if prefix and prefix not in Validator.get_all_prefixes():
                message = f"Node property 'id' has a value '{node}' with a CURIE prefix '{prefix}' is not represented in Biolink Model JSON-LD context"
                errors.append(ValidationError(node, error_type, message, MessageLevel.ERROR))
        return errors

    @staticmethod
    def validate_edge_property_values(subject: str, object: str, data: dict) -> list:
        """
        Validate an edge property's value.

        Parameters
        ----------
        subject: str
            Subject identifier
        object: str
            Object identifier
        data: dict
            Edge properties

        Returns
        -------
        list
            A list of errors for a given edge

        """
        errors = []
        error_type = ErrorType.INVALID_EDGE_PROPERTY_VALUE
        prefixes = Validator.get_all_prefixes()

        if PrefixManager.is_curie(subject):
            prefix = PrefixManager.get_prefix(subject)
            if prefix and prefix not in prefixes:
                message = f"Edge property 'subject' has a value '{subject}' with a CURIE prefix '{prefix}' that is not represented in Biolink Model JSON-LD context"
                errors.append(
                    ValidationError(f"{subject}-{object}", error_type, message, MessageLevel.ERROR)
                )
        else:
            message = f"Edge property 'subject' has a value '{subject}' which is not a proper CURIE"
            errors.append(
                ValidationError(f"{subject}-{object}", error_type, message, MessageLevel.ERROR)
            )

        if PrefixManager.is_curie(object):
            prefix = PrefixManager.get_prefix(object)
            if prefix not in prefixes:
                message = f"Edge property 'object' has a value '{object}' with a CURIE prefix '{prefix}' that is not represented in Biolink Model JSON-LD context"
                errors.append(
                    ValidationError(f"{subject}-{object}", error_type, message, MessageLevel.ERROR)
                )
        else:
            message = f"Edge property 'object' has a value '{object}' which is not a proper CURIE"
            errors.append(
                ValidationError(f"{subject}-{object}", error_type, message, MessageLevel.ERROR)
            )
        if 'relation' in data:
            if PrefixManager.is_curie(data['relation']):
                prefix = PrefixManager.get_prefix(data['relation'])
                if prefix not in prefixes:
                    message = f"Edge property 'relation' has a value '{data['relation']}' with a CURIE prefix '{prefix}' that is not represented in Biolink Model JSON-LD context"
                    errors.append(
                        ValidationError(
                            f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                        )
                    )
            else:
                message = f"Edge property 'relation' has a value '{data['relation']}' which is not a proper CURIE"
                errors.append(
                    ValidationError(f"{subject}-{object}", error_type, message, MessageLevel.ERROR)
                )
        return errors

    @staticmethod
    def validate_categories(node: str, data: dict) -> list:
        """
        Validate ``category`` field of a given node.

        Parameters
        ----------
        node: str
            Node identifier
        data: dict
            Node properties

        Returns
        -------
        list
            A list of errors for a given node

        """
        toolkit = get_toolkit()
        error_type = ErrorType.INVALID_CATEGORY
        errors = []
        categories = data.get('category')
        if categories is None:
            message = "Node does not have a 'category' property"
            errors.append(ValidationError(node, error_type, message, MessageLevel.ERROR))
        elif not isinstance(categories, list):
            message = f"Node property 'category' expected to be of type {list}"
            errors.append(ValidationError(node, error_type, message, MessageLevel.ERROR))
        else:
            for category in categories:
                if PrefixManager.is_curie(category):
                    category = PrefixManager.get_reference(category)
                m = re.match(r"^([A-Z][a-z\d]+)+$", category)
                if not m:
                    # category is not CamelCase
                    error_type = ErrorType.INVALID_CATEGORY
                    message = f"Category '{category}' is not in CamelCase form"
                    errors.append(ValidationError(node, error_type, message, MessageLevel.ERROR))
                formatted_category = camelcase_to_sentencecase(category)
                if not toolkit.is_category(formatted_category):
                    message = f"Category '{category}' not in Biolink Model"
                    errors.append(ValidationError(node, error_type, message, MessageLevel.ERROR))
                else:
                    c = toolkit.get_element(formatted_category.lower())
                    if c:
                        if category != c.name and category in c.aliases:
                            message = f"Category {category} is actually an alias for {c.name}; Should replace '{category}' with '{c.name}'"
                            errors.append(
                                ValidationError(node, error_type, message, MessageLevel.ERROR)
                            )
        return errors

    @staticmethod
    def validate_edge_predicate(subject: str, object: str, data: dict) -> list:
        """
        Validate ``edge_predicate`` field of a given edge.

        Parameters
        ----------
        subject: str
            Subject identifier
        object: str
            Object identifier
        data: dict
            Edge properties

        Returns
        -------
        list
            A list of errors for a given edge

        """
        toolkit = get_toolkit()
        error_type = ErrorType.INVALID_EDGE_PREDICATE
        errors = []
        edge_predicate = data.get('predicate')
        if edge_predicate is None:
            message = "Edge does not have an 'predicate' property"
            errors.append(
                ValidationError(f"{subject}-{object}", error_type, message, MessageLevel.ERROR)
            )
        elif not isinstance(edge_predicate, str):
            message = f"Edge property 'edge_predicate' expected to be of type 'string'"
            errors.append(
                ValidationError(f"{subject}-{object}", error_type, message, MessageLevel.ERROR)
            )
        else:
            if PrefixManager.is_curie(edge_predicate):
                edge_predicate = PrefixManager.get_reference(edge_predicate)
            m = re.match(r"^([a-z_][^A-Z\s]+_?[a-z_][^A-Z\s]+)+$", edge_predicate)
            if m:
                p = toolkit.get_element(snakecase_to_sentencecase(edge_predicate))
                if p is None:
                    message = f"Edge label '{edge_predicate}' not in Biolink Model"
                    errors.append(
                        ValidationError(
                            f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                        )
                    )
                elif edge_predicate != p.name and edge_predicate in p.aliases:
                    message = f"Edge label '{edge_predicate}' is actually an alias for {p.name}; Should replace {edge_predicate} with {p.name}"
                    errors.append(
                        ValidationError(
                            f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                        )
                    )
            else:
                message = f"Edge label '{edge_predicate}' is not in snake_case form"
                errors.append(
                    ValidationError(f"{subject}-{object}", error_type, message, MessageLevel.ERROR)
                )
        return errors

    @staticmethod
    def report(errors: List[ValidationError]) -> List:
        """
        Prepare error report.

        Parameters
        ----------
        errors: List[ValidationError]
            List of kgx.validator.ValidationError

        Returns
        -------
        List
            A list of formatted errors

        """
        return [str(x) for x in errors]

    @staticmethod
    def write_report(errors: List[ValidationError], outstream: TextIO) -> None:
        """
        Write error report to a file

        Parameters
        ----------
        errors: List[ValidationError]
            List of kgx.validator.ValidationError
        outstream: TextIO
            The stream to write to

        """
        for x in Validator.report(errors):
            outstream.write(f"{x}\n")



import json
from typing import Dict, List

from kgx.prefix_manager import PrefixManager
from kgx.graph.base_graph import BaseGraph

"""
Generate a knowledge map that corresponds to TRAPI KnowledgeMap.
Specification based on TRAPI Draft PR: https://github.com/NCATSTranslator/ReasonerAPI/pull/171

"""


def generate_knowledge_map(graph: BaseGraph, name: str, filename: str) -> None:
    """
    Generate a knowlege map that describes the composition of the graph
    and write to ``filename``.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    name: Optional[str]
        Name for the graph
    filename: str
        The file to write the knowledge map to

    """
    knowledge_map = summarize_graph(graph, name)
    WH = open(filename, 'w')
    json.dump(knowledge_map, WH, indent=4)


def summarize_graph(graph: BaseGraph, name: str = None, **kwargs) -> Dict:
    """
    Generate a knowlege map that describes the composition of the graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    name: Optional[str]
        Name for the graph
    kwargs: Dict
        Any additional arguments

    Returns
    -------
    Dict
        A knowledge map dictionary corresponding to the graph

    """
    node_stats = summarize_nodes(graph)
    edge_stats = summarize_edges(graph)
    graph_stats = {'knowledge_map': {'nodes': node_stats, 'edges': edge_stats}}
    if name:
        graph_stats['name'] = name
    return graph_stats


def summarize_nodes(graph: BaseGraph) -> Dict:
    """
    Summarize the nodes in a graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph

    Returns
    -------
    Dict
        The node stats

    """
    id_prefixes = set()
    count = 0
    count_by_source = {'unknown': 0}

    for n, data in graph.nodes(data=True):
        prefix = PrefixManager.get_prefix(n)
        count += 1
        if prefix not in id_prefixes:
            id_prefixes.add(prefix)
        if 'provided_by' in data:
            for s in data['provided_by']:
                if s in count_by_source:
                    count_by_source[s] += 1
                else:
                    count_by_source[s] = 1
        else:
            count_by_source['unknown'] += 1

    node_stats = {
        'id_prefixes': list(id_prefixes),
        'count': count,
        'count_by_source': count_by_source,
    }
    return node_stats


def summarize_edges(graph: BaseGraph) -> List[Dict]:
    """
    Summarize the edges in a graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph

    Returns
    -------
    List[Dict]
        The edge stats

    """
    association_map = {}
    for u, v, k, data in graph.edges(keys=True, data=True):
        if graph.has_node(u):
            subject_node = graph.nodes()[u]
            subject_category = subject_node['category'][0]
        else:
            subject_category = 'unknown'
        if graph.has_node(v):
            object_node = graph.nodes()[v]
            object_category = object_node['category'][0]
        else:
            object_category = 'unknown'

        triple = (subject_category, data['predicate'], object_category)
        if triple not in association_map:
            association_map[triple] = {
                'subject': triple[0],
                'predicate': triple[1],
                'object': triple[2],
                'relations': set(),
                'count': 0,
                'count_by_source': {'unknown': 0},
            }

        if data['relation'] not in association_map[triple]['relations']:
            association_map[triple]['relations'].add(data['relation'])

        association_map[triple]['count'] += 1
        if 'provided_by' in data:
            for s in data['provided_by']:
                if s not in association_map[triple]['count_by_source']:
                    association_map[triple]['count_by_source'][s] = 1
                else:
                    association_map[triple]['count_by_source'][s] += 1
        else:
            association_map[triple]['count_by_source']['unknown'] += 1

    edge_stats = []
    for k, v in association_map.items():
        kedge = v
        relations = list(v['relations'])
        kedge['relations'] = relations
        edge_stats.append(kedge)
    return edge_stats



import kgx
import click
from typing import List, Tuple, Optional, Set

from kgx.config import get_logger, get_config
from kgx.cli.cli_utils import (
    get_input_file_types,
    parse_source,
    apply_operations,
    graph_summary,
    validate,
    neo4j_download,
    neo4j_upload,
    transform,
    merge,
    summary_report_types,
)

log = get_logger()
config = get_config()


def error(msg):
    log.error(msg)
    quit()


@click.group()
@click.version_option(version=kgx.__version__, prog_name=kgx.__name__)
def cli():
    """
    Knowledge Graph Exchange CLI entrypoint.
    \f

    """
    pass


@cli.command(name='graph-summary')
@click.argument('inputs', required=True, type=click.Path(exists=True), nargs=-1)
@click.option(
    '--input-format',
    '-i',
    required=True,
    help=f'The input format. Can be one of {get_input_file_types()}',
)
@click.option('--input-compression', '-c', required=False, help='The input compression type')
@click.option('--output', '-o', required=True, type=click.Path(exists=False))
@click.option(
    '--report-type',
    '-r',
    required=False,
    type=str,
    help=f'The summary report type. Must be one of {tuple(summary_report_types.keys())}',
    default='kgx-map',
)
@click.option('--stream', '-s', is_flag=True, help='Parse input as a stream')
@click.option(
    '--node-facet-properties',
    required=False,
    multiple=True,
    help='A list of node properties from which to generate counts per value for those properties',
)
@click.option(
    '--edge-facet-properties',
    required=False,
    multiple=True,
    help='A list of edge properties from which to generate counts per value for those properties',
)
def graph_summary_wrapper(
    inputs: List[str],
    input_format: str,
    input_compression: str,
    output: str,
    report_type: str,
    stream: bool,
    node_facet_properties: Optional[Set],
    edge_facet_properties: Optional[Set],
):
    """
    Loads and summarizes a knowledge graph from a set of input files.
    \f

    Parameters
    ----------
    inputs: List[str]
        Input file
    input_format: str
        Input file format
    input_compression: str
        The input compression type
    output: str
        Where to write the output (stdout, by default)
    report_type: str
        The summary report type
    stream: bool
        Whether to parse input as a stream
    node_facet_properties: Optional[Set]
        A list of node properties from which to generate counts per value for those properties.
        For example, ``['provided_by']``
    edge_facet_properties: Optional[Set]
        A list of edge properties from which to generate counts per value for those properties.
        For example, ``['provided_by']``
    """
    graph_summary(
        inputs,
        input_format,
        input_compression,
        output,
        report_type,
        stream,
        node_facet_properties=list(node_facet_properties),
        edge_facet_properties=list(edge_facet_properties),
    )


@cli.command(name='validate')
@click.argument('inputs', required=True, type=click.Path(exists=True), nargs=-1)
@click.option(
    '--input-format',
    '-i',
    required=True,
    help=f'The input format. Can be one of {get_input_file_types()}',
)
@click.option('--input-compression', '-c', required=False, help='The input compression type')
@click.option(
    '--output',
    '-o',
    required=False,
    type=click.Path(exists=False),
    help='File to write validation reports to',
)
@click.option('--stream', '-s', is_flag=True, help='Parse input as a stream')
def validate_wrapper(
    inputs: List[str], input_format: str, input_compression: str, output: str, stream: bool
):
    """
    Run KGX validator on an input file to check for Biolink Model compliance.
    \f

    Parameters
    ----------
    inputs: List[str]
        Input files
    input_format: str
        The input format
    input_compression: str
        The input compression type
    output: str
        Path to output file
    stream: bool
        Whether to parse input as a stream

    """
    validate(inputs, input_format, input_compression, output, stream)


@cli.command(name='neo4j-download')
@click.option(
    '--uri',
    '-l',
    required=True,
    type=str,
    help='Neo4j URI to download from. For example, https://localhost:7474',
)
@click.option('--username', '-u', required=True, type=str, help='Neo4j username')
@click.option('--password', '-p', required=True, type=str, help='Neo4j password')
@click.option('--output', '-o', required=True, type=click.Path(exists=False), help='Output')
@click.option(
    '--output-format',
    '-f',
    required=True,
    help=f'The output format. Can be one of {get_input_file_types()}',
)
@click.option('--output-compression', '-d', required=False, help='The output compression type')
@click.option('--stream', '-s', is_flag=True, help='Parse input as a stream')
@click.option(
    '--node-filters',
    '-n',
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f'Filters for filtering nodes from the input graph',
)
@click.option(
    '--edge-filters',
    '-e',
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f'Filters for filtering edges from the input graph',
)
def neo4j_download_wrapper(
    uri: str,
    username: str,
    password: str,
    output: str,
    output_format: str,
    output_compression: str,
    stream: bool,
    node_filters: Tuple,
    edge_filters: Tuple,
):
    """
    Download nodes and edges from Neo4j database.
    \f

    Parameters
    ----------
    uri: str
        Neo4j URI. For example, https://localhost:7474
    username: str
        Username for authentication
    password: str
        Password for authentication
    output: str
        Where to write the output (stdout, by default)
    output_format: str
        The output type (``tsv``, by default)
    output_compression: str
        The output compression type
    stream: bool
        Whether to parse input as a stream
    node_filters: Tuple[str, str]
        Node filters
    edge_filters: Tuple[str, str]
        Edge filters

    """
    neo4j_download(
        uri,
        username,
        password,
        output,
        output_format,
        output_compression,
        stream,
        node_filters,
        edge_filters,
    )


@cli.command(name='neo4j-upload')
@click.argument('inputs', required=True, type=click.Path(exists=True), nargs=-1)
@click.option(
    '--input-format',
    '-i',
    required=True,
    help=f'The input format. Can be one of {get_input_file_types()}',
)
@click.option('--input-compression', '-c', required=False, help='The input compression type')
@click.option(
    '--uri',
    '-l',
    required=True,
    type=str,
    help='Neo4j URI to upload to. For example, https://localhost:7474',
)
@click.option('--username', '-u', required=True, type=str, help='Neo4j username')
@click.option('--password', '-p', required=True, type=str, help='Neo4j password')
@click.option('--stream', '-s', is_flag=True, help='Parse input as a stream')
@click.option(
    '--node-filters',
    '-n',
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f'Filters for filtering nodes from the input graph',
)
@click.option(
    '--edge-filters',
    '-e',
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f'Filters for filtering edges from the input graph',
)
def neo4j_upload_wrapper(
    inputs: List[str],
    input_format: str,
    input_compression: str,
    uri: str,
    username: str,
    password: str,
    stream: bool,
    node_filters: Tuple[str, str],
    edge_filters: Tuple[str, str],
):
    """
    Upload a set of nodes/edges to a Neo4j database.
    \f

    Parameters
    ----------
    inputs: List[str]
        A list of files that contains nodes/edges
    input_format: str
        The input format
    input_compression: str
        The input compression type
    uri: str
        The full HTTP address for Neo4j database
    username: str
        Username for authentication
    password: str
        Password for authentication
    stream: bool
        Whether to parse input as a stream
    node_filters: Tuple[str, str]
        Node filters
    edge_filters: Tuple[str, str]
        Edge filters

    """
    neo4j_upload(
        inputs,
        input_format,
        input_compression,
        uri,
        username,
        password,
        stream,
        node_filters,
        edge_filters,
    )


@cli.command(name='transform')
@click.argument('inputs', required=False, type=click.Path(exists=True), nargs=-1)
@click.option(
    '--input-format',
    '-i',
    required=False,
    help=f'The input format. Can be one of {get_input_file_types()}',
)
@click.option('--input-compression', '-c', required=False, help='The input compression type')
@click.option('--output', '-o', required=False, type=click.Path(exists=False), help='Output')
@click.option(
    '--output-format',
    '-f',
    required=False,
    help=f'The output format. Can be one of {get_input_file_types()}',
)
@click.option('--output-compression', '-d', required=False, help='The output compression type')
@click.option('--stream', is_flag=True, help='Parse input as a stream')
@click.option(
    '--node-filters',
    '-n',
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f'Filters for filtering nodes from the input graph',
)
@click.option(
    '--edge-filters',
    '-e',
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f'Filters for filtering edges from the input graph',
)
@click.option('--transform-config', required=False, type=str, help=f'Transform config YAML')
@click.option(
    '--source', required=False, type=str, multiple=True, help='Source(s) from the YAML to process'
)
@click.option('--processes', '-p', required=False, type=int, default=1, help='Number of processes to use')
def transform_wrapper(
    inputs: List[str],
    input_format: str,
    input_compression: str,
    output: str,
    output_format: str,
    output_compression: str,
    stream: bool,
    node_filters: Tuple,
    edge_filters: Tuple,
    transform_config: str,
    source: List,
    processes: int,
):
    """
    Transform a Knowledge Graph from one serialization form to another.
    \f

    Parameters
    ----------
    inputs: List[str]
        A list of files that contains nodes/edges
    input_format: str
        The input format
    input_compression: str
        The input compression type
    output: str
        The output file
    output_format: str
        The output format
    output_compression: str
        The output compression typ
    stream: bool
        Wheter or not to stream
    node_filters: Tuple[str, str]
        Node filters
    edge_filters: Tuple[str, str]
        Edge filters
    transform_config: str
        Transform config YAML
    source: List
        A list of source(s) to load from the YAML
    processes: int
        Number of processes to use

    """
    transform(
        inputs,
        input_format,
        input_compression,
        output,
        output_format,
        output_compression,
        stream,
        node_filters,
        edge_filters,
        transform_config,
        source,
        processes=processes,
    )


@cli.command(name='merge')
@click.option('--merge-config', required=True, type=str)
@click.option(
    '--source', required=False, type=str, multiple=True, help='Source(s) from the YAML to process'
)
@click.option(
    '--destination',
    required=False,
    type=str,
    multiple=True,
    help='Destination(s) from the YAML to process',
)
@click.option('--processes', '-p', required=False, type=int, default=1, help='Number of processes to use')
def merge_wrapper(merge_config: str, source: List, destination: List, processes: int):
    """
    Load nodes and edges from files and KGs, as defined in a config YAML, and merge them into a single graph.
    The merged graph can then be written to a local/remote Neo4j instance OR be serialized into a file.
    \f

    .. note::
        Everything here is driven by the ``merge-config`` YAML.

    Parameters
    ----------
    merge_config: str
        Merge config YAML
    source: List
        A list of source to load from the YAML
    destination: List
        A list of destination to write to, as defined in the YAML
    processes: int
        Number of processes to use

    """
    merge(merge_config, source, destination, processes)



import yaml
from typing import Dict, List, Optional

from kgx.graph.base_graph import BaseGraph

TOTAL_NODES = 'total_nodes'
NODE_CATEGORIES = 'node_categories'
COUNT_BY_CATEGORY = 'count_by_category'

TOTAL_EDGES = 'total_edges'
EDGE_PREDICATES = 'predicates'
COUNT_BY_EDGE_PREDICATES = 'count_by_predicates'
COUNT_BY_SPO = 'count_by_spo'

# Note: the format of the stats generated might change in the future


def generate_graph_stats(
    graph: BaseGraph,
    graph_name: str,
    filename: str,
    node_facet_properties: Optional[List] = None,
    edge_facet_properties: Optional[List] = None,
) -> None:
    """
    Generate stats from Graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    graph_name: str
        Name for the graph
    filename: str
        Filename to write the stats to
    node_facet_properties: Optional[List]
        A list of properties to facet on. For example, ``['provided_by']``
    edge_facet_properties: Optional[List]
        A list of properties to facet on. For example, ``['provided_by']``

    """
    stats = summarize_graph(graph, graph_name, node_facet_properties, edge_facet_properties)
    WH = open(filename, 'w')
    yaml.dump(stats, WH)


def summarize_graph(
    graph: BaseGraph,
    name: str = None,
    node_facet_properties: Optional[List] = None,
    edge_facet_properties: Optional[List] = None,
) -> Dict:
    """
    Summarize the entire graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    name: str
        Name for the graph
    node_facet_properties: Optional[List]
        A list of properties to facet on. For example, ``['provided_by']``
    edge_facet_properties: Optional[List]
        A list of properties to facet on. For example, ``['provided_by']``

    Returns
    -------
    Dict
        The stats dictionary

    """
    node_stats = summarize_nodes(graph, node_facet_properties)
    edge_stats = summarize_edges(graph, edge_facet_properties)
    stats = {
        'graph_name': name if name else graph.name,
        'node_stats': node_stats,
        'edge_stats': edge_stats,
    }
    return stats


def summarize_nodes(graph: BaseGraph, facet_properties: Optional[List] = None) -> Dict:
    """
    Summarize the nodes in a graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    facet_properties: Optional[List]
        A list of properties to facet on

    Returns
    -------
    Dict
        The node stats

    """
    stats: Dict = {
        TOTAL_NODES: 0,
        NODE_CATEGORIES: set(),
        COUNT_BY_CATEGORY: {'unknown': {'count': 0}},
    }

    stats[TOTAL_NODES] = len(graph.nodes())
    if facet_properties:
        for facet_property in facet_properties:
            stats[facet_property] = set()

    for n, data in graph.nodes(data=True):
        if 'category' not in data:
            stats[COUNT_BY_CATEGORY]['unknown']['count'] += 1
            continue
        categories = data['category']
        stats[NODE_CATEGORIES].update(categories)
        for category in categories:
            if category in stats[COUNT_BY_CATEGORY]:
                stats[COUNT_BY_CATEGORY][category]['count'] += 1
            else:
                stats[COUNT_BY_CATEGORY][category] = {'count': 1}

            if facet_properties:
                for facet_property in facet_properties:
                    stats = get_facet_counts(
                        data, stats, COUNT_BY_CATEGORY, category, facet_property
                    )

    stats[NODE_CATEGORIES] = sorted(list(stats[NODE_CATEGORIES]))
    if facet_properties:
        for facet_property in facet_properties:
            stats[facet_property] = sorted(list(stats[facet_property]))
    return stats


def summarize_edges(graph: BaseGraph, facet_properties: Optional[List] = None) -> Dict:
    """
    Summarize the edges in a graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    facet_properties: Optional[List]
        The properties to facet on

    Returns
    -------
    Dict
        The edge stats

    """
    stats: Dict = {
        TOTAL_EDGES: 0,
        EDGE_PREDICATES: set(),
        COUNT_BY_EDGE_PREDICATES: {'unknown': {'count': 0}},
        COUNT_BY_SPO: {},
    }

    stats[TOTAL_EDGES] = len(graph.edges())
    if facet_properties:
        for facet_property in facet_properties:
            stats[facet_property] = set()

    for u, v, k, data in graph.edges(keys=True, data=True):
        if 'predicate' not in data:
            stats[COUNT_BY_EDGE_PREDICATES]['unknown']['count'] += 1
            edge_predicate = "unknown"
        else:
            edge_predicate = data['predicate']
            stats[EDGE_PREDICATES].add(edge_predicate)
            if edge_predicate in stats[COUNT_BY_EDGE_PREDICATES]:
                stats[COUNT_BY_EDGE_PREDICATES][edge_predicate]['count'] += 1
            else:
                stats[COUNT_BY_EDGE_PREDICATES][edge_predicate] = {'count': 1}

            if facet_properties:
                for facet_property in facet_properties:
                    stats = get_facet_counts(
                        data, stats, COUNT_BY_EDGE_PREDICATES, edge_predicate, facet_property
                    )

        u_data = graph.nodes()[u]
        v_data = graph.nodes()[v]

        if 'category' in u_data:
            u_category = u_data['category'][0]
        else:
            u_category = "unknown"

        if 'category' in v_data:
            v_category = v_data['category'][0]
        else:
            v_category = "unknown"

        key = f"{u_category}-{edge_predicate}-{v_category}"
        if key in stats[COUNT_BY_SPO]:
            stats[COUNT_BY_SPO][key]['count'] += 1
        else:
            stats[COUNT_BY_SPO][key] = {'count': 1}

        if facet_properties:
            for facet_property in facet_properties:
                stats = get_facet_counts(data, stats, COUNT_BY_SPO, key, facet_property)

    stats[EDGE_PREDICATES] = sorted(list(stats[EDGE_PREDICATES]))
    if facet_properties:
        for facet_property in facet_properties:
            stats[facet_property] = sorted(list(stats[facet_property]))

    return stats


def get_facet_counts(data: Dict, stats: Dict, x: str, y: str, facet_property: str) -> Dict:
    """
    Facet on ``facet_property`` and record the count for ``stats[x][y][facet_property]``.

    Parameters
    ----------
    data: dict
        Node/edge data dictionary
    stats: dict
        The stats dictionary
    x: str
        first key
    y: str
        second key
    facet_property: str
        The property to facet on

    Returns
    -------
    Dict
        The stats dictionary

    """
    if facet_property in data:
        if isinstance(data[facet_property], list):
            for k in data[facet_property]:
                if facet_property not in stats[x][y]:
                    stats[x][y][facet_property] = {}

                if k in stats[x][y][facet_property]:
                    stats[x][y][facet_property][k]['count'] += 1
                else:
                    stats[x][y][facet_property][k] = {'count': 1}
                    stats[facet_property].update([k])
        else:
            k = data[facet_property]
            if facet_property not in stats[x][y]:
                stats[x][y][facet_property] = {}

            if k in stats[x][y][facet_property]:
                stats[x][y][facet_property][k]['count'] += 1
            else:
                stats[x][y][facet_property][k] = {'count': 1}
                stats[facet_property].update([k])
    else:
        if facet_property not in stats[x][y]:
            stats[x][y][facet_property] = {}
        if 'unknown' in stats[x][y][facet_property]:
            stats[x][y][facet_property]['unknown']['count'] += 1
        else:
            stats[x][y][facet_property]['unknown'] = {'count': 1}
            stats[facet_property].update(['unknown'])
    return stats



import gzip
import re
import jsonlines
from typing import Optional, Any, Generator

from kgx.source.json_source import JsonSource


class JsonlSource(JsonSource):
    """
    JsonlSource is responsible for reading data as records
    from JSON Lines.
    """

    def __init__(self):
        super().__init__()

    def parse(
        self,
        filename: str,
        format: str = 'jsonl',
        compression: Optional[str] = None,
        provided_by: Optional[str] = None,
        **kwargs: Any,
    ) -> Generator:
        """
        This method reads from JSON Lines and yields records.

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``json``)
        compression: Optional[str]
            The compression type (``gz``)
        provided_by: Optional[str]
            The name of the source providing the input file
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for records

        """
        if provided_by:
            self.graph_metadata['provided_by'] = [provided_by]
        if re.search(f'nodes.{format}', filename):
            m = self.read_node
        elif re.search(f'edges.{format}', filename):
            m = self.read_edge
        else:
            raise TypeError(f"Unrecognized file: {filename}")

        if compression == 'gz':
            with gzip.open(filename, 'rb') as FH:
                reader = jsonlines.Reader(FH)
                for obj in reader:
                    yield m(obj)
        else:
            with jsonlines.open(filename) as FH:
                for obj in FH:
                    yield m(obj)



import importlib
import logging
import sys
from os import path
import json
from typing import Dict, Any, Optional

import requests
import yaml

from kgx.graph.base_graph import BaseGraph

config: Optional[Dict[str, Any]] = None
logger: Optional[logging.Logger] = None
graph_store_class: Optional[BaseGraph] = None
jsonld_context_map: Dict = {}

CONFIG_FILENAME = path.join(path.dirname(path.abspath(__file__)), 'config.yml')


def get_config(filename: str = CONFIG_FILENAME) -> Dict:
    """
    Get config as a Dictionary

    Parameters
    ----------
    filename: str
        The filename with all the configuration

    Returns
    -------
    Dict
        A Dictionary containing all the entries from the config YAML

    """
    global config
    if config is None:
        config = yaml.load(open(filename), Loader=yaml.FullLoader)
    return config


def get_jsonld_context(name: str = "biolink"):
    """
    Get contents of a JSON-LD context.

    Returns
    -------
    Dict
        the contents of the JSON-LD context

    """
    content = None
    if name in jsonld_context_map:
        content = jsonld_context_map[name]
    else:
        filepath = config['jsonld-context'][name]  # type: ignore
        if filepath.startswith('http'):
            try:
                content = requests.get(filepath).json()
            except ConnectionError:
                raise Exception(f'Unable to download JSON-LD context from {filepath}')
        else:
            if path.exists(filepath):
                content = json.load(open(filepath))

        if '@context' in content:
            content = content['@context']
            jsonld_context_map[name] = content
    return content


def get_logger(name: str = 'KGX') -> logging.Logger:
    """
    Get an instance of logger.

    Parameters
    ----------
    name: str
        The name of logger

    Returns
    -------
    logging.Logger
        An instance of logging.Logger

    """
    global logger
    if logger is None:
        config = get_config()
        logger = logging.getLogger(name)
        handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(config['logging']['format'])
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(config['logging']['level'])
        logger.propagate = False
    return logger


def get_graph_store_class() -> Any:
    """
    Get a reference to the graph store class, as defined the config.
    Defaults to ``kgx.graph.nx_graph.NxGraph``

    Returns
    -------
    Any
        A reference to the graph store class

    """
    global graph_store_class
    if not graph_store_class:
        config = get_config()
        if 'graph_store' in config:
            name = config['graph_store']
        else:
            name = 'kgx.graph.nx_graph.NxGraph'
        module_name = '.'.join(name.split('.')[0:-1])
        class_name = name.split('.')[-1]
        graph_store_class = getattr(importlib.import_module(module_name), class_name)
    return graph_store_class



from collections import OrderedDict
from typing import List, Optional, Any, Union, Dict, Tuple
import rdflib
from linkml_model.meta import Element, SlotDefinition, ClassDefinition
from cachetools import cached, LRUCache
from rdflib import Namespace, URIRef
from rdflib.namespace import RDF, RDFS, OWL, SKOS

from kgx.config import get_logger
from kgx.prefix_manager import PrefixManager
from kgx.utils.graph_utils import get_category_via_superclass
from kgx.utils.kgx_utils import (
    get_curie_lookup_service,
    contract,
    get_cache,
    get_toolkit,
    sentencecase_to_snakecase,
    sentencecase_to_camelcase,
    get_biolink_ancestors,
)

log = get_logger()

OBAN = Namespace('http://purl.org/oban/')
BIOLINK = Namespace('https://w3id.org/biolink/vocab/')
OIO = Namespace('http://www.geneontology.org/formats/oboInOwl#')
OBO = Namespace('http://purl.obolibrary.org/obo/')

property_mapping: Dict = dict()
reverse_property_mapping: Dict = dict()

# TODO: this should be populated via bmt
is_property_multivalued = {
    'id': False,
    'subject': False,
    'object': False,
    'predicate': False,
    'description': False,
    'synonym': True,
    'in_taxon': False,
    'same_as': True,
    'name': False,
    'has_evidence': False,
    'provided_by': True,
    'category': True,
    'publications': True,
    'type': False,
    'relation': False,
}


top_level_terms = {
    OBO.term('CL_0000000'): 'cell',
    OBO.term('UBERON_0001062'): 'anatomical_entity',
    OBO.term('PATO_0000001'): 'quality',
    OBO.term('NCBITaxon_131567'): 'organism',
    OBO.term('CLO_0000031'): 'cell_line',
    OBO.term('MONDO_0000001'): 'disease',
    OBO.term('CHEBI_23367'): 'molecular_entity',
    OBO.term('CHEBI_23888'): 'drug',
    OBO.term('UPHENO_0001001'): 'phenotypic_feature',
    OBO.term('GO_0008150'): 'biological_process',
    OBO.term('GO_0009987'): 'cellular_process',
    OBO.term('GO_0005575'): 'cellular_component',
    OBO.term('GO_0003674'): 'molecular_function',
    OBO.term('SO_0000704'): 'gene',
    OBO.term('GENO_0000002'): 'variant_locus',
    OBO.term('GENO_0000536'): 'genotype',
    OBO.term('SO_0000110'): 'sequence_feature',
    OBO.term('ECO_0000000'): 'evidence',
    OBO.term('PW_0000001'): 'pathway',
    OBO.term('IAO_0000310'): 'publication',
    OBO.term('SO_0001483'): 'snv',
    OBO.term('GENO_0000871'): 'haplotype',
    OBO.term('SO_0001024'): 'haplotype',
    OBO.term('SO_0000340'): 'chromosome',
    OBO.term('SO_0000104'): 'protein',
    OBO.term('SO_0001500'): 'phenotypic_marker',
    OBO.term('SO_0000001'): 'region',
    OBO.term('HP_0032223'): 'blood_group',
    OBO.term('HP_0031797'): 'clinical_course',
    OBO.term('HP_0040279'): 'frequency',
    OBO.term('HP_0000118'): 'phenotypic_abnormality',
    OBO.term('HP_0032443'): 'past_medical_history',
    OBO.term('HP_0000005'): 'mode_of_inheritance',
    OBO.term('HP_0012823'): 'clinical_modifier',
}


def infer_category(iri: URIRef, rdfgraph: rdflib.Graph) -> Optional[List]:
    """
    Infer category for a given iri by traversing rdfgraph.

    Parameters
    ----------
    iri: rdflib.term.URIRef
        IRI
    rdfgraph: rdflib.Graph
        A graph to traverse

    Returns
    -------
    Optional[List]
        A list of category corresponding to the given IRI

    """
    closure = list(rdfgraph.transitive_objects(iri, URIRef(RDFS.subClassOf)))
    category = [top_level_terms[x] for x in closure if x in top_level_terms.keys()]
    if category:
        log.debug(
            "Inferred category as {} based on transitive closure over 'subClassOf' relation".format(
                category
            )
        )
    else:
        subj = closure[-1]
        if subj == iri:
            return category
        subject_curie: Optional[str] = contract(subj)
        if subject_curie and '_' in subject_curie:
            fixed_curie = subject_curie.split(':', 1)[1].split('_', 1)[1]
            log.warning("Malformed CURIE {} will be fixed to {}".format(subject_curie, fixed_curie))
            subject_curie = fixed_curie
        cls = get_curie_lookup_service()
        category = list(get_category_via_superclass(cls.ontology_graph, subject_curie))
    return category


@cached(LRUCache(maxsize=1024))
def get_biolink_element(prefix_manager: PrefixManager, predicate: Any) -> Optional[Element]:
    """
    Returns a Biolink Model element for a given predicate.

    Parameters
    ----------
    prefix_manager: PrefixManager
        An instance of prefix manager
    predicate: Any
        The CURIE of a predicate

    Returns
    -------
    Optional[Element]
        The corresponding Biolink Model element

    """
    toolkit = get_toolkit()
    if prefix_manager.is_iri(predicate):
        predicate_curie = prefix_manager.contract(predicate)
    else:
        predicate_curie = predicate
    if prefix_manager.is_curie(predicate_curie):
        reference = prefix_manager.get_reference(predicate_curie)
    else:
        reference = predicate_curie
    element = toolkit.get_element(reference)
    if not element:
        try:
            mapping = toolkit.get_element_by_mapping(predicate)
            if mapping:
                element = toolkit.get_element(mapping)
        except ValueError as e:
            log.error(e)
    return element


def process_predicate(
    prefix_manager: PrefixManager, p: Union[URIRef, str], predicate_mapping: Optional[Dict] = None
) -> Tuple:
    """
    Process a predicate where the method checks if there is a mapping in Biolink Model.

    Parameters
    ----------
    prefix_manager: PrefixManager
        An instance of prefix manager
    p: Union[URIRef, str]
        The predicate
    predicate_mapping: Optional[Dict]
        Predicate mappings

    Returns
    -------
    Tuple[str, str, str, str]
        A tuple that contains the Biolink CURIE (if available), the Biolink slot_uri CURIE (if available),
        the CURIE form of p, the reference of p

    """
    if prefix_manager.is_iri(p):
        predicate = prefix_manager.contract(str(p))
    else:
        predicate = None
    if prefix_manager.is_curie(p):
        property_name = prefix_manager.get_reference(p)
        predicate = p
    else:
        if predicate and prefix_manager.is_curie(predicate):
            property_name = prefix_manager.get_reference(predicate)
        else:
            property_name = p
            predicate = f":{p}"
    element = get_biolink_element(prefix_manager, p)
    canonical_uri = None
    if element:
        if isinstance(element, SlotDefinition):
            # predicate corresponds to a biolink slot
            if element.definition_uri:
                element_uri = prefix_manager.contract(element.definition_uri)
            else:
                element_uri = f"biolink:{sentencecase_to_snakecase(element.name)}"
            if element.slot_uri:
                canonical_uri = element.slot_uri
        elif isinstance(element, ClassDefinition):
            # this will happen only when the IRI is actually
            # a reference to a class
            element_uri = prefix_manager.contract(element.class_uri)
        else:
            element_uri = f"biolink:{sentencecase_to_camelcase(element.name)}"
        if 'biolink:Attribute' in get_biolink_ancestors(element.name):
            element_uri = f"biolink:{sentencecase_to_snakecase(element.name)}"
        if not predicate:
            predicate = element_uri
    else:
        # no mapping to biolink model;
        # look at predicate mappings
        element_uri = None
        if predicate_mapping:
            if p in predicate_mapping:
                property_name = predicate_mapping[p]
                predicate = f":{property_name}"
        # cache[p] = {'element_uri': element_uri, 'canonical_uri': canonical_uri,
        # 'predicate': predicate, 'property_name': property_name}
    return element_uri, canonical_uri, predicate, property_name



import gzip
from itertools import chain
from typing import Optional, Dict, Generator, Any
import ijson
import stringcase
from bmt import Toolkit

from kgx.prefix_manager import PrefixManager
from kgx.config import get_logger
from kgx.source.json_source import JsonSource
from kgx.utils.kgx_utils import get_biolink_element, format_biolink_slots

log = get_logger()


class ObographSource(JsonSource):
    """
    ObographSource is responsible for reading data as records
    from an OBO Graph JSON.
    """

    HAS_OBO_NAMESPACE = 'http://www.geneontology.org/formats/oboInOwl#hasOBONamespace'
    SKOS_EXACT_MATCH = 'http://www.w3.org/2004/02/skos/core#exactMatch'

    def __init__(self):
        super().__init__()
        self.toolkit = Toolkit()
        self.ecache: Dict = {}

    def parse(
        self,
        filename: str,
        format: str = 'json',
        compression: Optional[str] = None,
        provided_by: Optional[str] = None,
        **kwargs: Any,
    ) -> Generator:
        """
        This method reads from JSON and yields records.

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``json``)
        compression: Optional[str]
            The compression type (``gz``)
        provided_by: Optional[str]
            The name of the source providing the input file
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for records

        """
        if provided_by:
            self.graph_metadata['provided_by'] = [provided_by]
        n = self.read_nodes(filename, compression)
        e = self.read_edges(filename, compression)
        yield from chain(n, e)

    def read_nodes(self, filename: str, compression: Optional[str] = None) -> Generator:
        """
        Read node records from a JSON.

        Parameters
        ----------
        filename: str
            The filename to read from
        compression: Optional[str]
            The compression type

        Returns
        -------
        Generator
            A generator for node records

        """
        if compression and compression == 'gz':
            FH = gzip.open(filename, 'rb')
        else:
            FH = open(filename, 'rb')
        for n in ijson.items(FH, 'graphs.item.nodes.item'):
            yield self.read_node(n)

    def read_node(self, node: Dict) -> Dict:
        """
        Read and parse a node record.

        Parameters
        ----------
        node: Dict
            The node record

        Returns
        -------
        Dict
            The processed node

        """
        curie = self.prefix_manager.contract(node['id'])
        node_properties = {}
        if 'meta' in node:
            node_properties = self.parse_meta(node['id'], node['meta'])

        fixed_node = dict()
        fixed_node['id'] = curie
        if 'lbl' in node:
            fixed_node['name'] = node['lbl']
        fixed_node['iri'] = node['id']

        if 'description' in node_properties:
            fixed_node['description'] = node_properties['description']
        if 'synonym' in node_properties:
            fixed_node['synonym'] = node_properties['synonym']
        if 'xrefs' in node_properties:
            fixed_node['xref'] = node_properties['xrefs']
        if 'subsets' in node_properties:
            fixed_node['subsets'] = node_properties['subsets']

        if 'category' not in node:
            category = self.get_category(curie, node)
            if category:
                fixed_node['category'] = [category]
            else:
                fixed_node['category'] = ['biolink:OntologyClass']
        if 'equivalent_nodes' in node_properties:
            equivalent_nodes = node_properties['equivalent_nodes']
            fixed_node['same_as'] = equivalent_nodes
            # for n in node_properties['equivalent_nodes']:
            #     data = {'subject': fixed_node['id'], 'predicate': 'biolink:same_as', 'object': n, 'relation': 'owl:sameAs'}
            #     super().load_node({'id': n, 'category': ['biolink:OntologyClass']})
            #     self.graph.add_edge(fixed_node['id'], n, **data)
        return super().read_node(fixed_node)

    def read_edges(self, filename: str, compression: Optional[str] = None) -> Generator:
        """
        Read edge records from a JSON.

        Parameters
        ----------
        filename: str
            The filename to read from
        compression: Optional[str]
            The compression type

        Returns
        -------
        Generator
            A generator for edge records

        """
        if compression == 'gz':
            FH = gzip.open(filename, 'rb')
        else:
            FH = open(filename, 'rb')
        for e in ijson.items(FH, 'graphs.item.edges.item'):
            yield self.read_edge(e)

    def read_edge(self, edge: Dict) -> Dict:
        """
        Read and parse an edge record.

        Parameters
        ----------
        edge: Dict
            The edge record

        Returns
        -------
        Dict
            The processed edge

        """
        fixed_edge = dict()
        fixed_edge['subject'] = self.prefix_manager.contract(edge['sub'])
        if PrefixManager.is_iri(edge['pred']):
            curie = self.prefix_manager.contract(edge['pred'])
            if curie in self.ecache:
                edge_predicate = self.ecache[curie]
            else:
                element = get_biolink_element(curie)
                if not element:
                    try:
                        mapping = self.toolkit.get_element_by_mapping(edge['pred'])
                        if mapping:
                            element = self.toolkit.get_element(mapping)
                    except ValueError as e:
                        log.error(e)

                if element:
                    edge_predicate = format_biolink_slots(element.name.replace(',', ''))
                    fixed_edge['predicate'] = edge_predicate
                else:
                    edge_predicate = 'biolink:related_to'
                self.ecache[curie] = edge_predicate
            fixed_edge['predicate'] = edge_predicate
            fixed_edge['relation'] = curie
        else:
            if edge['pred'] == 'is_a':
                fixed_edge['predicate'] = 'biolink:subclass_of'
                fixed_edge['relation'] = 'rdfs:subClassOf'
            elif edge['pred'] == 'has_part':
                fixed_edge['predicate'] = 'biolink:has_part'
                fixed_edge['relation'] = "BFO:0000051"
            elif edge['pred'] == 'part_of':
                fixed_edge['predicate'] = 'biolink:part_of'
                fixed_edge['relation'] = "BFO:0000050"
            else:
                fixed_edge['predicate'] = f"biolink:{edge['pred'].replace(' ', '_')}"
                fixed_edge['relation'] = edge['pred']

        fixed_edge['object'] = self.prefix_manager.contract(edge['obj'])
        for x in edge.keys():
            if x not in {'sub', 'pred', 'obj'}:
                fixed_edge[x] = edge[x]
        return super().read_edge(fixed_edge)

    def get_category(self, curie: str, node: dict) -> Optional[str]:
        """
        Get category for a given CURIE.

        Parameters
        ----------
        curie: str
            Curie for node
        node: dict
            Node data

        Returns
        -------
        Optional[str]
            Category for the given node CURIE.

        """
        category = None
        # use meta.basicPropertyValues
        if 'meta' in node and 'basicPropertyValues' in node['meta']:
            for p in node['meta']['basicPropertyValues']:
                if p['pred'] == self.HAS_OBO_NAMESPACE:
                    category = p['val']
                    element = self.toolkit.get_element(category)
                    if element:
                        category = (
                            f"biolink:{stringcase.pascalcase(stringcase.snakecase(element.name))}"
                        )
                    else:
                        element = self.toolkit.get_element_by_mapping(category)
                        if element:
                            category = f"biolink:{stringcase.pascalcase(stringcase.snakecase(element.name))}"
                        else:
                            category = 'biolink:OntologyClass'

        if not category or category == 'biolink:OntologyClass':
            prefix = PrefixManager.get_prefix(curie)
            # TODO: the mapping should be via biolink-model lookups
            if prefix == 'HP':
                category = "biolink:PhenotypicFeature"
            elif prefix == 'CHEBI':
                category = "biolink:ChemicalSubstance"
            elif prefix == 'MONDO':
                category = "biolink:Disease"
            elif prefix == 'UBERON':
                category = "biolink:AnatomicalEntity"
            elif prefix == 'SO':
                category = "biolink:SequenceFeature"
            elif prefix == 'CL':
                category = "biolink:Cell"
            elif prefix == 'PR':
                category = "biolink:Protein"
            elif prefix == 'NCBITaxon':
                category = "biolink:OrganismalEntity"
            else:
                log.debug(
                    f"{curie} Could not find a category mapping for '{category}'; Defaulting to 'biolink:OntologyClass'"
                )
        return category

    def parse_meta(self, node: str, meta: Dict) -> Dict:
        """
        Parse 'meta' field of a node.

        Parameters
        ----------
        node: str
            Node identifier
        meta: Dict
            meta dictionary for the node

        Returns
        -------
        Dict
            A dictionary that contains 'description', 'synonyms',
            'xrefs', and 'equivalent_nodes'.

        """
        # cross species links are in meta; this needs to be parsed properly too
        # do not put assumptions in code; import as much as possible

        properties = {}
        if 'definition' in meta:
            # parse 'definition' as 'description'
            description = meta['definition']['val']
            properties['description'] = description

        if 'subsets' in meta:
            # parse 'subsets'
            subsets = meta['subsets']
            properties['subsets'] = [x.split('#')[1] if '#' in x else x for x in subsets]

        if 'synonyms' in meta:
            # parse 'synonyms' as 'synonym'
            synonyms = [s['val'] for s in meta['synonyms']]
            properties['synonym'] = synonyms

        if 'xrefs' in meta:
            # parse 'xrefs' as 'xrefs'
            xrefs = [x['val'] for x in meta['xrefs']]
            properties['xrefs'] = xrefs

        if 'deprecated' in meta:
            # parse 'deprecated' flag
            properties['deprecated'] = meta['deprecated']

        equivalent_nodes = []
        if 'basicPropertyValues' in meta:
            # parse SKOS_EXACT_MATCH entries as 'equivalent_nodes'
            for p in meta['basicPropertyValues']:
                if p['pred'] in {self.SKOS_EXACT_MATCH}:
                    n = self.prefix_manager.contract(p['val'])
                    if not n:
                        n = p['val']
                    equivalent_nodes.append(n)
        properties['equivalent_nodes'] = equivalent_nodes
        return properties



import gzip
import ijson
from itertools import chain
from typing import Dict, Tuple, Generator, Optional, Any

from kgx.source.json_source import JsonSource


# TODO: update for TRAPI 1.0 spec


class TrapiSource(JsonSource):
    """
    TrapiSource is responsible for reading data as records
    from a TRAPI JSON.
    """

    def __init__(self):
        super().__init__()
        self._node_properties = set()
        self._edge_properties = set()

    def parse(
        self,
        filename: str,
        format: str = 'json',
        compression: Optional[str] = None,
        provided_by: Optional[str] = None,
        **kwargs: Any
    ) -> Generator:
        """
        This method reads from a JSON and yields records.

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``trapi-json``)
        compression: Optional[str]
            The compression type (``gz``)
        provided_by: Optional[str]
            The name of the source providing the input file
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for node and edge records

        """
        if provided_by:
            self.graph_metadata['provided_by'] = [provided_by]
        n = self.read_nodes(filename, compression)
        e = self.read_edges(filename, compression)
        yield from chain(n, e)

    def read_nodes(self, filename: str, compression: Optional[str] = None) -> Generator:
        """
        Read node records from a JSON.

        Parameters
        ----------
        filename: str
            The filename to read from
        compression: Optional[str]
            The compression type

        Returns
        -------
        Generator
            A generator for node records

        """
        if compression == 'gz':
            FH = gzip.open(filename, 'rb')
        else:
            FH = open(filename, 'rb')
        for n in ijson.items(FH, 'knowledge_graph.nodes.item'):
            yield self.load_node(n)

    def read_edges(self, filename: str, compression: Optional[str] = None) -> Generator:
        """
        Read edge records from a JSON.

        Parameters
        ----------
        filename: str
            The filename to read from
        compression: Optional[str]
            The compression type

        Returns
        -------
        Generator
            A generator for edge records

        """
        if compression == 'gz':
            FH = gzip.open(filename, 'rb')
        else:
            FH = open(filename, 'rb')
        for e in ijson.items(FH, 'knowledge_graph.edges.item'):
            yield self.load_edge(e)

    def load_node(self, node: Dict) -> Tuple[str, Dict]:
        """
        Load a node into an instance of BaseGraph

        .. Note::
            This method transformers Reasoner Std API format fields to Biolink Model fields.

        Parameters
        ----------
        node : Dict
            A node

        """

        if 'type' in node and 'category' not in node:
            node['category'] = node['type']
            del node['type']
        return super().read_node(node)

    def load_edge(self, edge: Dict) -> Tuple[str, str, str, Dict]:
        """
        Load an edge into an instance of BaseGraph

        .. Note::
            This methods transformers Reasoner Std API format fields to Biolink Model fields.

        Parameters
        ----------
        edge : Dict
            An edge

        """
        if 'source_id' in edge:
            edge['subject'] = edge['source_id']
        if 'target_id' in edge:
            edge['object'] = edge['target_id']
        if 'relation_label' in edge:
            edge['predicate'] = edge['relation_label'][0]
        return super().read_edge(edge)



import gzip
import re

import pandas as pd
from typing import Optional, Generator, Any, Dict, Tuple

import yaml

from kgx.prefix_manager import PrefixManager
from kgx.config import get_logger
from kgx.source import Source
from kgx.utils.kgx_utils import (
    validate_node,
    sanitize_import,
    validate_edge,
    generate_uuid,
    generate_edge_key,
)
from kgx.utils.rdf_utils import process_predicate

log = get_logger()

SSSOM_NODE_PROPERTY_MAPPING = {
    'subject_id': 'id',
    'subject_category': 'category',
    'object_id': 'id',
    'object_category': 'category',
}


class SssomSource(Source):
    """
    SssomSource is responsible for reading data as records
    from an SSSOM file.
    """

    def __init__(self):
        super().__init__()
        self.predicate_mapping = {}

    def set_prefix_map(self, m: Dict) -> None:
        """
        Add or override default prefix to IRI map.

        Parameters
        ----------
        m: Dict
            Prefix to IRI map

        """
        self.prefix_manager.set_prefix_map(m)

    def set_reverse_prefix_map(self, m: Dict) -> None:
        """
        Add or override default IRI to prefix map.

        Parameters
        ----------
        m: Dict
            IRI to prefix map

        """
        self.prefix_manager.set_reverse_prefix_map(m)

    def parse(
        self,
        filename: str,
        format: str,
        compression: Optional[str] = None,
        provided_by: str = None,
        **kwargs: Any,
    ) -> Generator:
        """
        Parse a SSSOM TSV

        Parameters
        ----------
        filename: str
            File to read from
        format: str
            The input file format (``tsv``, by default)
        compression: Optional[str]
            The compression (``gz``)
        provided_by: Optional[str]
            Define the source providing the input file
        kwargs: Dict
            Any additional arguments

        Returns
        -------
        Generator
            A generator for node and edge records

        """
        if 'delimiter' not in kwargs:
            kwargs['delimiter'] = '\t'
        self.parse_header(filename, compression)
        if provided_by:
            self.graph_metadata['provided_by'] = [provided_by]
        if 'mapping_provider' in self.graph_metadata:
            self.graph_metadata['provided_by'] = self.graph_metadata['mapping_provider']
        if compression:
            FH = gzip.open(filename, 'rb')
        else:
            FH = open(filename)
        file_iter = pd.read_csv(
            FH,
            comment='#',
            dtype=str,
            chunksize=10000,
            low_memory=False,
            keep_default_na=False,
            **kwargs,
        )
        for chunk in file_iter:
            yield from self.load_edges(chunk)

    def parse_header(self, filename: str, compression: Optional[str] = None) -> None:
        """
        Parse metadata from SSSOM headers.

        Parameters
        ----------
        filename: str
            Filename to parse
        compression: Optional[str]
            Compression type

        """
        yamlstr = ""
        if compression:
            FH = gzip.open(filename, 'rb')
        else:
            FH = open(filename)
        for line in FH:
            if line.startswith('#'):
                yamlstr += re.sub('^#', '', line)
            else:
                break
        if yamlstr:
            metadata = yaml.safe_load(yamlstr)
            log.info(f"Metadata: {metadata}")
            if 'curie_map' in metadata:
                self.prefix_manager.update_prefix_map(metadata['curie_map'])
            for k, v in metadata.items():
                self.graph_metadata[k] = v

    def load_node(self, node: Dict) -> Tuple[str, Dict]:
        """
        Load a node into an instance of BaseGraph

        Parameters
        ----------
        node: Dict
            A node

        Returns
        -------
        Optional[Tuple[str, Dict]]
            A tuple that contains node id and node data

        """
        node = validate_node(node)
        node_data = sanitize_import(node.copy())
        if 'id' in node_data:
            n = node_data['id']
            if 'provided_by' in self.graph_metadata and 'provided_by' not in node_data.keys():
                node_data['provided_by'] = self.graph_metadata['provided_by']
            self.node_properties.update(list(node_data.keys()))
            return n, node_data
        else:
            log.info("Ignoring node with no 'id': {}".format(node))

    def load_edges(self, df: pd.DataFrame) -> Generator:
        """
        Load edges from pandas.DataFrame into an instance of BaseGraph

        Parameters
        ----------
        df : pandas.DataFrame
            Dataframe containing records that represent edges

        Returns
        -------
        Generator
            A generator for edge records

        """
        for obj in df.to_dict('records'):
            yield from self.load_edge(obj)

    def load_edge(self, edge: Dict) -> Generator:
        """
        Load an edge into an instance of BaseGraph

        Parameters
        ----------
        edge : Dict
            An edge

        Returns
        -------
        Generator
            A generator for node and edge records

        """
        (element_uri, canonical_uri, predicate, property_name) = process_predicate(
            self.prefix_manager, edge['predicate_id'], self.predicate_mapping
        )
        if element_uri:
            edge_predicate = element_uri
        elif predicate:
            edge_predicate = predicate
        else:
            edge_predicate = property_name
        if canonical_uri:
            edge_predicate = element_uri
        data = {
            'subject': edge['subject_id'],
            'predicate': edge_predicate,
            'object': edge['object_id'],
        }
        del edge['predicate_id']
        data = validate_edge(data)
        subject_node = {}
        object_node = {}
        for k, v in edge.items():
            if k in SSSOM_NODE_PROPERTY_MAPPING:
                if k.startswith('subject'):
                    mapped_k = SSSOM_NODE_PROPERTY_MAPPING[k]
                    if mapped_k == 'category' and not PrefixManager.is_curie(v):
                        v = f"biolink:OntologyClass"
                    subject_node[mapped_k] = v
                elif k.startswith('object'):
                    mapped_k = SSSOM_NODE_PROPERTY_MAPPING[k]
                    if mapped_k == 'category' and not PrefixManager.is_curie(v):
                        v = f"biolink:OntologyClass"
                    object_node[mapped_k] = v
                else:
                    log.info(f"Ignoring {k} {v}")
            else:
                data[k] = v

        objs = [self.load_node(subject_node), self.load_node(object_node)]

        for k, v in self.graph_metadata.items():
            if k not in {'curie_map'}:
                data[k] = v

        edge_data = sanitize_import(data.copy())
        if 'subject' in edge_data and 'object' in edge_data:
            if 'id' not in edge_data:
                edge_data['id'] = generate_uuid()
            s = edge_data['subject']
            o = edge_data['object']
            if 'provided_by' in self.graph_metadata and 'provided_by' not in edge_data.keys():
                edge_data['provided_by'] = self.graph_metadata['provided_by']
            key = generate_edge_key(s, edge_data['predicate'], o)
            self.edge_properties.update(list(edge_data.keys()))
            objs.append((s, o, key, edge_data))
        else:
            log.info(
                "Ignoring edge with either a missing 'subject' or 'object': {}".format(edge_data)
            )

        for o in objs:
            yield o



import itertools
import os
from typing import Dict, Union, Generator, List

from kgx.config import get_logger
from kgx.sink import GraphSink, Sink, TsvSink, JsonSink, JsonlSink, NeoSink, RdfSink
from kgx.source import (
    GraphSource,
    Source,
    TsvSource,
    JsonSource,
    JsonlSource,
    ObographSource,
    TrapiSource,
    NeoSource,
    RdfSource,
)
from kgx.source.sssom_source import SssomSource
from kgx.source.owl_source import OwlSource
from kgx.utils.kgx_utils import apply_graph_operations

SOURCE_MAP = {
    'tsv': TsvSource,
    'csv': TsvSource,
    'graph': GraphSource,
    'json': JsonSource,
    'jsonl': JsonlSource,
    'obojson': ObographSource,
    'obo-json': ObographSource,
    'trapi-json': TrapiSource,
    'neo4j': NeoSource,
    'nt': RdfSource,
    'owl': OwlSource,
    'sssom': SssomSource,
}

SINK_MAP = {
    'tsv': TsvSink,
    'csv': TsvSink,
    'graph': GraphSink,
    'json': JsonSink,
    'jsonl': JsonlSink,
    'neo4j': NeoSink,
    'nt': RdfSink,
}


log = get_logger()


class Transformer(object):
    """
    The Transformer class is responsible for transforming data from one
    form to another.

    Parameters
    ----------
    stream: bool
        Whether or not to stream

    """

    def __init__(self, stream: bool = False):
        self.stream = stream
        self.node_filters = {}
        self.edge_filters = {}
        self.store = self.get_source('graph')
        self._seen_nodes = set()

    def transform(self, input_args: Dict, output_args: Dict = None) -> None:
        """
        Transform an input source and write to an output sink.

        If ``output_args`` is not defined then the data is persisted to
        an in-memory graph.

        Parameters
        ----------
        input_args: Dict
            Arguments relevant to your input source
        output_args: Dict
            Arguments relevant to your output sink

        """
        sources = []
        generators = []
        input_format = input_args['format']
        prefix_map = input_args.pop('prefix_map', {})
        predicate_mappings = input_args.pop('predicate_mappings', {})
        node_property_predicates = input_args.pop('node_property_predicates', {})
        node_filters = input_args.pop('node_filters', {})
        edge_filters = input_args.pop('edge_filters', {})
        operations = input_args.pop('operations', [])

        if input_format in {'neo4j', 'graph'}:
            source = self.get_source(input_format)
            source.set_prefix_map(prefix_map)
            source.set_node_filters(node_filters)
            self.node_filters = source.node_filters
            self.edge_filters = source.edge_filters
            source.set_edge_filters(edge_filters)
            self.node_filters = source.node_filters
            self.edge_filters = source.edge_filters
            if 'provided_by' not in input_args:
                if 'name' in input_args:
                    input_args['provided_by'] = input_args['name']
                else:
                    if 'uri' in input_args:
                        input_args['provided_by'] = input_args['uri']
            g = source.parse(**input_args)
            sources.append(source)
            generators.append(g)
        else:
            filename = input_args.pop('filename', {})
            for f in filename:
                source = self.get_source(input_format)
                source.set_prefix_map(prefix_map)
                if isinstance(source, RdfSource):
                    source.set_predicate_mapping(predicate_mappings)
                    source.set_node_property_predicates(node_property_predicates)
                source.set_node_filters(node_filters)
                self.node_filters = source.node_filters
                self.edge_filters = source.edge_filters
                source.set_edge_filters(edge_filters)
                self.node_filters = source.node_filters
                self.edge_filters = source.edge_filters
                if 'provided_by' not in input_args:
                    if 'name' in input_args:
                        input_args['provided_by'] = input_args.pop('name')
                    else:
                        input_args['provided_by'] = os.path.basename(f)
                g = source.parse(f, **input_args)
                sources.append(source)
                generators.append(g)

        source_generator = itertools.chain(*generators)

        if output_args:
            if self.stream:
                if output_args['format'] in {'tsv', 'csv'}:
                    if 'node_properties' not in output_args:
                        log.warning(
                            f"'node_properties' not defined for output while streaming. The exported {output_args['format']} will be limited to a subset of the columns."
                        )
                    if 'edge_properties' not in output_args:
                        log.warning(
                            f"'edge_properties' not defined for output while streaming. The exported {output_args['format']} will be limited to a subset of the columns."
                        )
                sink = self.get_sink(**output_args)
                if 'reverse_prefix_map' in output_args:
                    sink.set_reverse_prefix_map(output_args['reverse_prefix_map'])
                if isinstance(sink, RdfSink):
                    if 'reverse_predicate_mapping' in output_args:
                        sink.set_reverse_predicate_mapping(output_args['reverse_predicate_mapping'])
                    if 'property_types' in output_args:
                        sink.set_property_types(output_args['property_types'])
                # stream from source to sink
                self.process(source_generator, sink)
                sink.finalize()
            else:
                # stream from source to intermediate
                intermediate_sink = GraphSink(self.store.graph)
                intermediate_sink.node_properties.update(self.store.node_properties)
                intermediate_sink.edge_properties.update(self.store.edge_properties)
                self.process(source_generator, intermediate_sink)
                for s in sources:
                    intermediate_sink.node_properties.update(s.node_properties)
                    intermediate_sink.edge_properties.update(s.edge_properties)
                apply_graph_operations(intermediate_sink.graph, operations)
                # stream from intermediate to output sink
                intermediate_source = self.get_source('graph')
                intermediate_source.node_properties.update(intermediate_sink.node_properties)
                intermediate_source.edge_properties.update(intermediate_sink.edge_properties)
                intermediate_source_generator = intermediate_source.parse(intermediate_sink.graph)

                if output_args['format'] in {'tsv', 'csv'}:
                    if 'node_properties' not in output_args:
                        output_args['node_properties'] = intermediate_source.node_properties
                    if 'edge_properties' not in output_args:
                        output_args['edge_properties'] = intermediate_source.edge_properties
                    sink = self.get_sink(**output_args)
                    if 'reverse_prefix_map' in output_args:
                        sink.set_reverse_prefix_map(output_args['reverse_prefix_map'])
                    if isinstance(sink, RdfSink):
                        if 'reverse_predicate_mapping' in output_args:
                            sink.set_reverse_predicate_mapping(
                                output_args['reverse_predicate_mapping']
                            )
                    if 'property_types' in output_args:
                        sink.set_property_types(output_args['property_types'])
                else:
                    sink = self.get_sink(**output_args)
                    sink.node_properties.update(intermediate_source.node_properties)
                    sink.edge_properties.update(intermediate_source.edge_properties)

                self.process(intermediate_source_generator, sink)
                sink.finalize()
                self.store.node_properties.update(sink.node_properties)
                self.store.edge_properties.update(sink.edge_properties)
        else:
            # stream from source to intermediate
            sink = GraphSink(self.store.graph)
            self.process(source_generator, sink)
            sink.node_properties.update(self.store.node_properties)
            sink.edge_properties.update(self.store.edge_properties)
            for s in sources:
                sink.node_properties.update(s.node_properties)
                sink.edge_properties.update(s.edge_properties)
            sink.finalize()
            self.store.node_properties.update(sink.node_properties)
            self.store.edge_properties.update(sink.edge_properties)
            apply_graph_operations(sink.graph, operations)

    def process(self, source: Generator, sink: Sink) -> None:
        """
        This method is responsible for reading from ``source``
        and writing to ``sink`` by calling the relevant methods
        based on the incoming data.

        .. note::
            The streamed data must not be mutated.

        Parameters
        ----------
        source: Generator
            A generator from a Source
        sink: kgx.sink.sink.Sink
            An instance of Sink

        """
        for rec in source:
            if rec:
                if len(rec) == 4:
                    write_edge = True
                    if 'subject_category' in self.edge_filters:
                        if rec[0] in self._seen_nodes:
                            write_edge = True
                        else:
                            write_edge = False
                    if 'object_category' in self.edge_filters:
                        if rec[1] in self._seen_nodes:
                            if 'subject_category' in self.edge_filters:
                                if write_edge:
                                    write_edge = True
                            else:
                                write_edge = True
                        else:
                            write_edge = False
                    if write_edge:
                        sink.write_edge(rec[-1])
                else:
                    if 'category' in self.node_filters:
                        self._seen_nodes.add(rec[0])
                    sink.write_node(rec[-1])

    def save(self, output_args: Dict) -> None:
        """
        Save data from the in-memory store to a desired sink.

        Parameters
        ----------
        output_args: Dict
            Arguments relevant to your output sink

        """
        if not self.store:
            raise Exception("self.store is empty.")
        source = self.store
        source.node_properties.update(self.store.node_properties)
        source.edge_properties.update(self.store.edge_properties)
        source_generator = source.parse(self.store.graph)
        if 'node_properties' not in output_args:
            output_args['node_properties'] = source.node_properties
        if 'edge_properties' not in output_args:
            output_args['edge_properties'] = source.edge_properties
        sink = self.get_sink(**output_args)
        sink.node_properties.update(source.node_properties)
        sink.edge_properties.update(source.edge_properties)
        if 'reverse_prefix_map' in output_args:
            sink.set_reverse_prefix_map(output_args['reverse_prefix_map'])
        if isinstance(sink, RdfSink):
            if 'reverse_predicate_mapping' in output_args:
                sink.set_reverse_predicate_mapping(output_args['reverse_predicate_mapping'])
            if 'property_types' in output_args:
                sink.set_property_types(output_args['property_types'])
        self.process(source_generator, sink)
        sink.finalize()

    def get_source(self, format: str) -> Source:
        """
        Get an instance of Source that corresponds to a given format.

        Parameters
        ----------
        format: str
            The input store format

        Returns
        -------
        Source:
            An instance of kgx.source.Source

        """
        if format in SOURCE_MAP:
            s = SOURCE_MAP[format]
            return s()
        else:
            raise TypeError(f"{format} in an unrecognized format")

    def get_sink(self, **kwargs: Dict) -> Sink:
        """
        Get an instance of Sink that corresponds to a given format.

        Parameters
        ----------
        kwargs: Dict
            Arguments required for initializing an instance of Sink

        Returns
        -------
        Sink:
            An instance of kgx.sink.Sink

        """
        if kwargs['format'] in SINK_MAP:
            s = SINK_MAP[kwargs['format']]
            return s(**kwargs)
        else:
            raise TypeError(f"{kwargs['format']} in an unrecognized format")



import re
import tarfile
from typing import Dict, Tuple, Any, Generator, Optional, Union
import pandas as pd

from kgx.config import get_logger
from kgx.source.source import Source
from kgx.utils.kgx_utils import (
    generate_uuid,
    generate_edge_key,
    extension_types,
    archive_read_mode,
    remove_null,
    sanitize_import,
    validate_edge,
    validate_node,
)

log = get_logger()


class TsvSource(Source):
    """
    TsvSource is responsible for reading data as records
    from a TSV/CSV.
    """

    def __init__(self):
        super().__init__()

    def set_prefix_map(self, m: Dict) -> None:
        """
        Add or override default prefix to IRI map.

        Parameters
        ----------
        m: Dict
            Prefix to IRI map

        """
        self.prefix_manager.set_prefix_map(m)

    def set_reverse_prefix_map(self, m: Dict) -> None:
        """
        Add or override default IRI to prefix map.

        Parameters
        ----------
        m: Dict
            IRI to prefix map

        """
        self.prefix_manager.set_reverse_prefix_map(m)

    def parse(
        self,
        filename: str,
        format: str,
        compression: Optional[str] = None,
        provided_by: str = None,
        **kwargs: Any,
    ) -> Generator:
        """
        This method reads from a TSV/CSV and yields records.

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``tsv``, ``csv``)
        compression: Optional[str]
            The compression type (``tar``, ``tar.gz``)
        provided_by: Optional[str]
            The name of the source providing the input file
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for node and edge records

        """
        if 'delimiter' not in kwargs:
            # infer delimiter from file format
            kwargs['delimiter'] = extension_types[format]  # type: ignore
        if 'lineterminator' not in kwargs:
            # set '\n' to be the default line terminator to prevent
            # truncation of lines due to hidden/escaped carriage returns
            kwargs['lineterminator'] = '\n'  # type: ignore

        mode = archive_read_mode[compression] if compression in archive_read_mode else None

        if provided_by:
            self.graph_metadata['provided_by'] = [provided_by]
        if format == 'tsv':
            kwargs['quoting'] = 3  # type: ignore
        if mode:
            with tarfile.open(filename, mode=mode) as tar:
                for member in tar.getmembers():
                    f = tar.extractfile(member)
                    file_iter = pd.read_csv(
                        f,
                        dtype=str,
                        chunksize=10000,
                        low_memory=False,
                        keep_default_na=False,
                        **kwargs,
                    )
                    if re.search(f'nodes.{format}', member.name):
                        for chunk in file_iter:
                            self.node_properties.update(chunk.columns)
                            yield from self.read_nodes(chunk)
                    elif re.search(f'edges.{format}', member.name):
                        for chunk in file_iter:
                            self.edge_properties.update(chunk.columns)
                            yield from self.read_edges(chunk)
                    else:
                        raise Exception(f'Tar archive contains an unrecognized file: {member.name}')
        else:
            file_iter = pd.read_csv(
                filename,
                dtype=str,
                chunksize=10000,
                low_memory=False,
                keep_default_na=False,
                **kwargs,
            )
            if re.search(f'nodes.{format}', filename):
                for chunk in file_iter:
                    self.node_properties.update(chunk.columns)
                    yield from self.read_nodes(chunk)
            elif re.search(f'edges.{format}', filename):
                for chunk in file_iter:
                    self.edge_properties.update(chunk.columns)
                    yield from self.read_edges(chunk)
            else:
                raise Exception(f'Unrecognized file: {filename}')

    def read_nodes(self, df: pd.DataFrame) -> Generator:
        """
        Read records from pandas.DataFrame and yield records.

        Parameters
        ----------
        df: pandas.DataFrame
            Dataframe containing records that represent nodes

        Returns
        -------
        Generator
            A generator for node records

        """
        for obj in df.to_dict('records'):
            yield self.read_node(obj)

    def read_node(self, node: Dict) -> Optional[Tuple[str, Dict]]:
        """
        Prepare a node.

        Parameters
        ----------
        node: Dict
            A node

        Returns
        -------
        Optional[Tuple[str, Dict]]
            A tuple that contains node id and node data

        """
        node = validate_node(node)
        node_data = sanitize_import(node.copy())
        if 'id' in node_data:
            n = node_data['id']
            if 'provided_by' in self.graph_metadata and 'provided_by' not in node_data.keys():
                node_data['provided_by'] = self.graph_metadata['provided_by']
            self.node_properties.update(list(node_data.keys()))
            if self.check_node_filter(node_data):
                self.node_properties.update(node_data.keys())
                return n, node_data
        else:
            log.info(f"Ignoring node with no 'id': {node}")

    def read_edges(self, df: pd.DataFrame) -> Generator:
        """
        Load edges from pandas.DataFrame into an instance of BaseGraph.

        Parameters
        ----------
        df: pandas.DataFrame
            Dataframe containing records that represent edges

        Returns
        -------
        Generator
            A generator for edge records

        """
        for obj in df.to_dict('records'):
            yield self.read_edge(obj)

    def read_edge(self, edge: Dict) -> Optional[Tuple]:
        """
        Load an edge into an instance of BaseGraph.

        Parameters
        ----------
        edge: Dict
            An edge

        Returns
        -------
        Optional[Tuple]
            A tuple that contains subject id, object id, edge key, and edge data

        """
        edge = validate_edge(edge)
        edge_data = sanitize_import(edge.copy())
        if 'id' not in edge_data:
            edge_data['id'] = generate_uuid()
        s = edge_data['subject']
        o = edge_data['object']
        if 'provided_by' in self.graph_metadata and 'provided_by' not in edge_data.keys():
            edge_data['provided_by'] = self.graph_metadata['provided_by']
        key = generate_edge_key(s, edge_data['predicate'], o)
        self.edge_properties.update(list(edge_data.keys()))
        if self.check_edge_filter(edge_data):
            self.node_properties.update(edge_data.keys())
            return s, o, key, edge_data
